/*
 * armboot - Startup Code for OMAP3530/ARM Cortex CPU-core
 *
 * Copyright (c) 2004	Texas Instruments <r-woodruff2@ti.com>
 *
 * Copyright (c) 2001	Marius Gröger <mag@sysgo.de>
 * Copyright (c) 2002	Alex Züpke <azu@sysgo.de>
 * Copyright (c) 2002	Gary Jennejohn <garyj@denx.de>
 * Copyright (c) 2003	Richard Woodruff <r-woodruff2@ti.com>
 * Copyright (c) 2003	Kshitij <kshitij@ti.com>
 * Copyright (c) 2006-2008 Syed Mohammed Khasim <x0khasim@ti.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */

#include <asm-offsets.h>
#include <config.h>
#include <version.h>
#include <asm/system.h>
#include <linux/linkage.h>

.globl _start
_start: b	reset
	ldr	pc, _undefined_instruction
	ldr	pc, _software_interrupt
	ldr	pc, _prefetch_abort
	ldr	pc, _data_abort
	ldr	pc, _not_used
	ldr	pc, _irq
	ldr	pc, _fiq
#ifdef CONFIG_SPL_BUILD
_undefined_instruction: .word _undefined_instruction
_software_interrupt:	.word _software_interrupt
_prefetch_abort:	.word _prefetch_abort
_data_abort:		.word _data_abort
_not_used:		.word _not_used
_irq:			.word _irq
_fiq:			.word _fiq
_pad:			.word 0x12345678 /* now 16*4=64 */
#else
_undefined_instruction: .word undefined_instruction
_software_interrupt:	.word software_interrupt
_prefetch_abort:	.word prefetch_abort
_data_abort:		.word data_abort
_not_used:		.word not_used
_irq:			.word irq
_fiq:			.word fiq
_pad:			.word 0x12345678 /* now 16*4=64 */
#endif	/* CONFIG_SPL_BUILD */

.global _end_vect
_end_vect:

	.balignl 16,0xdeadbeef
/*************************************************************************
 *
 * Startup Code (reset vector)
 *
 * do important init only if we don't start from memory!
 * setup Memory and board specific bits prior to relocation.
 * relocate armboot to ram
 * setup stack
 *
 *************************************************************************/

.globl _TEXT_BASE
_TEXT_BASE:
	.word	CONFIG_SYS_TEXT_BASE

/*
 * These are defined in the board-specific linker script.
 */
.globl _bss_start_ofs
_bss_start_ofs:
	.word __bss_start - _start

.global	_image_copy_end_ofs
_image_copy_end_ofs:
	.word 	__image_copy_end - _start

.globl _bss_end_ofs
_bss_end_ofs:
	.word __bss_end__ - _start

.globl _end_ofs
_end_ofs:
	.word _end - _start

#ifdef CONFIG_USE_IRQ
/* IRQ stack memory (calculated at run-time) */
.globl IRQ_STACK_START
IRQ_STACK_START:
	.word	0x0badc0de

/* IRQ stack memory (calculated at run-time) */
.globl FIQ_STACK_START
FIQ_STACK_START:
	.word 0x0badc0de
#endif

/* IRQ stack memory (calculated at run-time) + 8 bytes */
.globl IRQ_STACK_START_IN
IRQ_STACK_START_IN:
	.word	0x0badc0de

/*
 * the actual reset code
 */

reset:
#ifdef CONFIG_XCODE
	bl xcode_init
#endif
	bl	save_boot_params
	/*
	 * set the cpu to SVC32 mode
	 */
	mrs	r0, cpsr
	bic	r0, r0, #0x1f
	orr	r0, r0, #0xd3
	msr	cpsr,r0

/*
 * Setup vector:
 * (OMAP4 spl TEXT_BASE is not 32 byte aligned.
 * Continue to use ROM code vector only in OMAP4 spl)
 */
#if !((defined(CONFIG_OMAP44XX) || defined(CONFIG_XCODE)) && defined(CONFIG_SPL_BUILD))
	/* Set V=0 in CP15 SCTRL register - for VBAR to point to vector */
	mrc	p15, 0, r0, c1, c0, 0	@ Read CP15 SCTRL Register
	bic	r0, #CR_V		@ V = 0
	mcr	p15, 0, r0, c1, c0, 0	@ Write CP15 SCTRL Register

	/* Set vector address in CP15 VBAR register */
	ldr	r0, =_start
	mcr	p15, 0, r0, c12, c0, 0	@Set VBAR
#endif

	/* the mask ROM code should have PLL and others stable */
#ifndef CONFIG_SKIP_LOWLEVEL_INIT
	bl	cpu_init_cp15
	bl	cpu_init_crit
#endif

/* Set stackpointer in internal RAM to call board_init_f */
call_board_init_f:
	ldr	sp, =(CONFIG_SYS_INIT_SP_ADDR)
	bic	sp, sp, #7 /* 8-byte alignment for ABI compliance */
	ldr	r0,=0x00000000
	bl	board_init_f

/*------------------------------------------------------------------------------*/

#ifndef CONFIG_SPL_BUILD
/*
 * void relocate_code (addr_sp, gd, addr_moni)
 *
 * This "function" does not return, instead it continues in RAM
 * after relocating the monitor code.
 *
 */
ENTRY(relocate_code)
	mov	r4, r0	/* save addr_sp */
	mov	r5, r1	/* save addr of gd */
	mov	r6, r2	/* save addr of destination */

	/* Set up the stack						    */
stack_setup:
	mov	sp, r4

	adr	r0, _start
	cmp	r0, r6
	moveq	r9, #0		/* no relocation. relocation offset(r9) = 0 */
	beq	clear_bss		/* skip relocation */
	mov	r1, r6			/* r1 <- scratch for copy_loop */
	ldr	r3, _image_copy_end_ofs
	add	r2, r0, r3		/* r2 <- source end address	    */

copy_loop:
	ldmia	r0!, {r9-r10}		/* copy from source address [r0]    */
	stmia	r1!, {r9-r10}		/* copy to   target address [r1]    */
	cmp	r0, r2			/* until source end address [r2]    */
	blo	copy_loop

	/*
	 * fix .rel.dyn relocations
	 */
	ldr	r0, _TEXT_BASE		/* r0 <- Text base */
	sub	r9, r6, r0		/* r9 <- relocation offset */
	ldr	r10, _dynsym_start_ofs	/* r10 <- sym table ofs */
	add	r10, r10, r0		/* r10 <- sym table in FLASH */
	ldr	r2, _rel_dyn_start_ofs	/* r2 <- rel dyn start ofs */
	add	r2, r2, r0		/* r2 <- rel dyn start in FLASH */
	ldr	r3, _rel_dyn_end_ofs	/* r3 <- rel dyn end ofs */
	add	r3, r3, r0		/* r3 <- rel dyn end in FLASH */
fixloop:
	ldr	r0, [r2]		/* r0 <- location to fix up, IN FLASH! */
	add	r0, r0, r9		/* r0 <- location to fix up in RAM */
	ldr	r1, [r2, #4]
	and	r7, r1, #0xff
	cmp	r7, #23			/* relative fixup? */
	beq	fixrel
	cmp	r7, #2			/* absolute fixup? */
	beq	fixabs
	/* ignore unknown type of fixup */
	b	fixnext
fixabs:
	/* absolute fix: set location to (offset) symbol value */
	mov	r1, r1, LSR #4		/* r1 <- symbol index in .dynsym */
	add	r1, r10, r1		/* r1 <- address of symbol in table */
	ldr	r1, [r1, #4]		/* r1 <- symbol value */
	add	r1, r1, r9		/* r1 <- relocated sym addr */
	b	fixnext
fixrel:
	/* relative fix: increase location by offset */
	ldr	r1, [r0]
	add	r1, r1, r9
fixnext:
	str	r1, [r0]
	add	r2, r2, #8		/* each rel.dyn entry is 8 bytes */
	cmp	r2, r3
	blo	fixloop
	b	clear_bss
_rel_dyn_start_ofs:
	.word __rel_dyn_start - _start
_rel_dyn_end_ofs:
	.word __rel_dyn_end - _start
_dynsym_start_ofs:
	.word __dynsym_start - _start

clear_bss:
	ldr	r0, _bss_start_ofs
	ldr	r1, _bss_end_ofs
	mov	r4, r6			/* reloc addr */
	add	r0, r0, r4
	add	r1, r1, r4
	mov	r2, #0x00000000		/* clear			    */

clbss_l:cmp	r0, r1			/* clear loop... */
	bhs	clbss_e			/* if reached end of bss, exit */
	str	r2, [r0]
	add	r0, r0, #4
	b	clbss_l
clbss_e:

/*
 * We are done. Do not return, instead branch to second part of board
 * initialization, now running from RAM.
 */
jump_2_ram:
/*
 * If I-cache is enabled invalidate it
 */
#ifndef CONFIG_SYS_ICACHE_OFF
	mcr	p15, 0, r0, c7, c5, 0	@ invalidate icache
	mcr     p15, 0, r0, c7, c10, 4	@ DSB
	mcr     p15, 0, r0, c7, c5, 4	@ ISB
#endif
/*
 * Move vector table
 */
#if !defined(CONFIG_TEGRA20)
	/* Set vector address in CP15 VBAR register */
	ldr     r0, =_start
	add     r0, r0, r9
	mcr     p15, 0, r0, c12, c0, 0  @Set VBAR
#endif /* !Tegra20 */

	ldr	r0, _board_init_r_ofs
	adr	r1, _start
	add	lr, r0, r1
	add	lr, lr, r9
	/* setup parameters for board_init_r */
	mov	r0, r5		/* gd_t */
	mov	r1, r6		/* dest_addr */
	/* jump to it ... */
	mov	pc, lr

_board_init_r_ofs:
	.word board_init_r - _start
ENDPROC(relocate_code)
#endif

/*************************************************************************
 *
 * void save_boot_params(u32 r0, u32 r1, u32 r2, u32 r3)
 *	__attribute__((weak));
 *
 * Stack pointer is not yet initialized at this moment
 * Don't save anything to stack even if compiled with -O0
 *
 *************************************************************************/
ENTRY(save_boot_params)
	bx	lr			@ back to my caller
ENDPROC(save_boot_params)
	.weak	save_boot_params

/*************************************************************************
 *
 * cpu_init_cp15
 *
 * Setup CP15 registers (cache, MMU, TLBs). The I-cache is turned on unless
 * CONFIG_SYS_ICACHE_OFF is defined.
 *
 *************************************************************************/
ENTRY(cpu_init_cp15)
	/*
	 * Invalidate L1 I/D
	 */
	mov	r0, #0			@ set up for MCR
	mcr	p15, 0, r0, c8, c7, 0	@ invalidate TLBs
	mcr	p15, 0, r0, c7, c5, 0	@ invalidate icache
	mcr	p15, 0, r0, c7, c5, 6	@ invalidate BP array
	mcr     p15, 0, r0, c7, c10, 4	@ DSB
	mcr     p15, 0, r0, c7, c5, 4	@ ISB

	/*
	 * disable MMU stuff and caches
	 */
	mrc	p15, 0, r0, c1, c0, 0
	bic	r0, r0, #0x00002000	@ clear bits 13 (--V-)
	bic	r0, r0, #0x00000007	@ clear bits 2:0 (-CAM)
	#orr	r0, r0, #0x00000002	@ set bit 1 (--A-) Align
	bic	r0, r0, #0x00000002	@ clear bit 1 (--A-) Align
	orr	r0, r0, #0x00000800	@ set bit 11 (Z---) BTB
#ifdef CONFIG_SYS_ICACHE_OFF
	bic	r0, r0, #0x00001000	@ clear bit 12 (I) I-cache
#else
	orr	r0, r0, #0x00001000	@ set bit 12 (I) I-cache
#endif
	mcr	p15, 0, r0, c1, c0, 0
	mov	pc, lr			@ back to my caller
ENDPROC(cpu_init_cp15)

#ifndef CONFIG_SKIP_LOWLEVEL_INIT
/*************************************************************************
 *
 * CPU_init_critical registers
 *
 * setup important registers
 * setup memory timing
 *
 *************************************************************************/
ENTRY(cpu_init_crit)
	/*
	 * Jump to board specific initialization...
	 * The Mask ROM will have already initialized
	 * basic memory. Go here to bump up clock rate and handle
	 * wake up conditions.
	 */
	b	lowlevel_init		@ go setup pll,mux,memory
ENDPROC(cpu_init_crit)
#endif

#ifdef CONFIG_XCODE
#define ARM(x...) x
#define THUMB(x...)

ENTRY(xcode_init)

		.macro disable_l2cc tmp1, tmp2
		ldr	\tmp1, =0xfff02000	
		mov	\tmp2, #0
		str \tmp2, [\tmp1, #0x100]
		.endm

		.macro disable_mmu_caches tmp1
		mrc p15, 0, \tmp1, c1, c0, 0 @Disable MMU, I/D cache
		bic	\tmp1, \tmp1, #(1<<12)
		bic	\tmp1, \tmp1, #(1<<11)
		bic \tmp1, \tmp1, #(1<<2)
		bic \tmp1, \tmp1, #1
		mcr p15, 0, \tmp1, c1, c0, 0
		.endm

		.macro inv_l2_cache tmp1, tmp2, tmp3
		ldr	\tmp1, =0xfff02000		@Invalidate all
		mov	\tmp2, #1
		lsl \tmp2, \tmp2, #16
		sub \tmp2, \tmp2, #1
		str \tmp2, [\tmp1, #0x77c]
20001:
		ldr \tmp3, [\tmp1, #0x77c]  @cache_wait_way()
		cmp \tmp2, \tmp3
		bne 20001b

		mov	\tmp2, #0				@cache_sync()
		str \tmp2, [\tmp1, #0x730]
		.endm

		.macro inv_l1_cache tmp1, tmp2, tmp3, tmp4, tmp5, tmp6, tmp7, tmp8, tmp9, tmp10
        @Invalidate L1 Instruction Cache, also flushes BTAC

        mov    \tmp1, #0                      @ sbz
        mcr    p15, 0, \tmp1, c7, c5, 0       @ iciallu - invalidate entire i cache, and flushes branch target cache

        @ invalidate data/unified caches

        mrc     p15, 1, \tmp1, c0, c0, 1      @ read clidr
        ands    \tmp4, \tmp1, #0x07000000        @ extract coherency level
        mov     \tmp4, \tmp4, lsr #23            @ total cache levels << 1
        beq     10005f                   @ if 0, no need to clean

        mov     \tmp9, #0                    @ \tmp9 holds current cache level << 1
10001:  
		add     \tmp3, \tmp9, \tmp9, lsr #1       @ \tmp3 holds cache "set" position
        mov     \tmp2, \tmp1, lsr \tmp3             @ bottom 3 bits are the cache-type for this level
        and     \tmp2, \tmp2, #7                 @ isolate those lower 3 bits
        cmp     \tmp2, #2
        blt     10004f                       @ no cache or only instruction cache at this level

        mcr     p15, 2, \tmp9, c0, c0, 0     @ write the cache size selection register
        isb                                @ isb to sync the change to the cachesizeid reg
        mrc     p15, 1, \tmp2, c0, c0, 0      @ reads current cache size id register
        and     \tmp3, \tmp2, #7                 @ extract the line length field
        add     \tmp3, \tmp3, #4                 @ add 4 for the line length offset (log2 16 bytes)
        ldr     \tmp5, =0x3ff
        ands    \tmp5, \tmp5, \tmp2, lsr #3         @ \tmp5 is the max number on the way size (right aligned)
        clz     \tmp6, \tmp5                     @ \tmp6 is the bit position of the way size increment
        ldr     \tmp7, =0x7fff
        ands    \tmp7, \tmp7, \tmp2, lsr #13        @ \tmp7 is the max number of the index size (right aligned)

10002:
	   mov     \tmp8, \tmp5                     @ \tmp8 working copy of the max way size (right aligned)

10003:
	   ARM(orr     \tmp10, \tmp9, \tmp8, lsl \tmp6)       @ factor in the way number and cache number into \tmp10
	   THUMB(mov 	\tmp10, \tmp8, lsl \tmp6)
	   THUMB(orr	\tmp10, \tmp9, \tmp10)
       ARM(orr     \tmp10, \tmp10, \tmp7, lsl \tmp3)       @ factor in the set number
	   THUMB(mov 	\tmp2, \tmp7, lsl \tmp3)
	   THUMB(orr	\tmp10, \tmp10, \tmp2)
       mcr     p15, 0, \tmp10, c7, c6, 2     @ invalidate by set/way
       subs    \tmp8, \tmp8, #1                 @ decrement the way number
       bge     10003b
       subs    \tmp7, \tmp7, #1                 @ decrement the set number
       bge     10002b
10004:
	   add     \tmp9, \tmp9, #2               @ increment the cache number
       cmp     \tmp4, \tmp9
       bgt     10001b

10005:
		.endm

		.macro enable_scu tmp1, tmp2
		mrc p15, 4, \tmp1, c15, c0, 0
		ldr \tmp2, [\tmp1, #0]
		orr \tmp2, \tmp2, #1
		str \tmp2, [\tmp1, #0]
		.endm

		.macro join_smp arg1, arg2, tmp1
		and \arg1, \arg1, #0x3
		mov	\arg1, \arg1, lsl #2
		and \arg2, \arg2, #0xf
		mov	\arg2, \arg2, lsl \arg1 
		mrc p15, 4, \tmp1, c15, c0, 0
		str	\arg2, [\tmp1, #0xc]
	
		mrc p15, 0, \arg1, c1, c0, 1 @Join SMP
		orr	\arg1, \arg1, #0x40
		mcr p15, 0, \arg1, c1, c0, 1

		mrc p15, 0, \arg1, c1, c0, 1 @enableMaintenanceBroadcast
		mov \arg2, \arg1
		orr	\arg1, \arg1, #0x1
		cmp \arg1, \arg2
		mcrne p15, 0, \arg1, c1, c0, 1
		.endm

		.macro enable_gic_proc_interface tmp1, tmp2
		mrc p15, 4, \tmp1, c15, c0, 0 @enableGICProcessorInterface
	
		ldr \tmp2, [\tmp1, #0x100] 
		orr \tmp2, \tmp2, #0x3
		str \tmp2, [\tmp1, #0x100]
		.endm

		.macro gic_init tmp1, tmp2
		mrc p15, 4, \tmp1, c15, c0, 0 @enableGIC
		add \tmp1, \tmp1, #0x1000
	
		ldr \tmp2, [\tmp1]
		orr \tmp2, \tmp2, #0x1
		str \tmp2, [\tmp1]
	
		enable_gic_proc_interface \tmp1, \tmp2
		.endm

		.macro set_priority_mask arg1, tmp1
		mrc p15, 4, \tmp1, c15, c0, 0
		str \arg1, [\tmp1, #0x0104]
		.endm

		.macro enable_int_id arg1, tmp1, tmp2, tmp3
		mov \tmp1, \arg1
		mrc p15, 4, \arg1, c15, c0, 0
		mov \tmp2, \tmp1
		mov \tmp2, \tmp2, lsr #5
		mov	\tmp2, \tmp2, lsl #2
		and	\tmp1, \tmp1, #0x1f
		mov	\tmp3, #1
		mov	\tmp3, \tmp3, lsl \tmp1
		and \tmp2, \tmp2, #0x1100
		str \tmp3, [\arg1, \tmp2]
		.endm

		.macro set_int_priority arg1, arg2, tmp1, tmp2, tmp3
		mov	\tmp1, \arg1
		mrc p15, 4, \arg1, c15, c0, 0
		and \arg2, \arg2, #0x1f
		mov \arg2, \arg2, lsl #3
		bic \tmp2, \tmp1, #0x3
		add \tmp2, \tmp2, #0x1400
		add \arg1, \arg1, \tmp2
		and \tmp1, \tmp1, #0x3
		mov \tmp1, \tmp1, lsl #3
		mov	\tmp3, #0xff
		mov \tmp3, \tmp3, lsl \tmp1
		mov \arg2, \arg2, lsl \tmp1
		ldr \tmp2, [\arg1]
		bic \tmp2, \tmp2, \tmp3
		orr \tmp2, \tmp2, \arg2
		str \tmp2, [\arg1]
		.endm

		disable_mmu_caches r0
		disable_l2cc r0, r1

		@Invalidate entire Unified TLB
		mov	r0, #0
		mcr p15, 0, r0, c8, c7, 0

		inv_l1_cache r0, r1, r2, r3, r4, r5, r7, r9, r10, r11

        mov     r0, #0
        mcr     p15, 0, r0, c7, c5, 6      @ bpiall - invalidate entire branch predictor array

		/* Write a marker here -- We reach here! */
		@ldr	r0, =0x0f000000
		@ldr	r1, =0xbeef0002
		@str r1, [r0]

		@ldr	r0, =0xffee0000
		@ldr	r1, =0xcafe0002
		@str r1, [r0]

		mrc p15, 0, r3, c0, c0, 5 @ We stop secondary CPU here
		ands r3, r3, #0x3
		beq primary

		@Secondary CPU setup
		enable_gic_proc_interface r0, r1
		mov r0, #0x1f
		set_priority_mask r0, r1
		mov r0, #0
		enable_int_id r0, r1, r2, r3
		mov r0, #0
		mov r1, #0
		set_int_priority r0, r1, r2, r3, r12
		mrc p15, 0, r0, c0, c0, 5
		ands r0, r0, #0x3
		mov	r1, #0xf
		join_smp r0, r1, r2

		mov r1, #0xff000000
		add r1, r1, #0x00ee0000
		mov r0, #0x1
		str r0, [r1]

hold:
		ldr r1, =0xffee0000  @temporary as gic wakeup doesn't working now.
		ldr r0, [r1]
		ldr r1, =0xcafebabe
		cmp r0, r1
		bne hold

loop:
		ldr r1, =0xffee0000
		ldr r0, =0xbabecafe
		str r0, [r1]

		b   loop

primary:
		@enable_scu r0, r1
		@mov r0, #0
		@mov r1, #0xf
		@join_smp r0, r1, r2
		gic_init r0, r1

		mov r12, lr 		@Backup return address

#if 0
		mov r0, #'U'
		bl printch
#endif

		mov	lr, r12
		mov	pc, lr			@ back to my caller
ENDPROC(xcode_init)

ENTRY(printch)
		ldr r1, =0xffeef400
		str r0, [r1]
		mov pc, lr
ENDPROC(printch)
#endif
#ifndef CONFIG_SPL_BUILD
/*
 *************************************************************************
 *
 * Interrupt handling
 *
 *************************************************************************
 */
@
@ IRQ stack frame.
@
#define S_FRAME_SIZE	72

#define S_OLD_R0	68
#define S_PSR		64
#define S_PC		60
#define S_LR		56
#define S_SP		52

#define S_IP		48
#define S_FP		44
#define S_R10		40
#define S_R9		36
#define S_R8		32
#define S_R7		28
#define S_R6		24
#define S_R5		20
#define S_R4		16
#define S_R3		12
#define S_R2		8
#define S_R1		4
#define S_R0		0

#define MODE_SVC 0x13
#define I_BIT	 0x80

/*
 * use bad_save_user_regs for abort/prefetch/undef/swi ...
 * use irq_save_user_regs / irq_restore_user_regs for IRQ/FIQ handling
 */

	.macro	bad_save_user_regs
	sub	sp, sp, #S_FRAME_SIZE		@ carve out a frame on current
						@ user stack
	stmia	sp, {r0 - r12}			@ Save user registers (now in
						@ svc mode) r0-r12
	ldr	r2, IRQ_STACK_START_IN		@ set base 2 words into abort
						@ stack
	ldmia	r2, {r2 - r3}			@ get values for "aborted" pc
						@ and cpsr (into parm regs)
	add	r0, sp, #S_FRAME_SIZE		@ grab pointer to old stack

	add	r5, sp, #S_SP
	mov	r1, lr
	stmia	r5, {r0 - r3}			@ save sp_SVC, lr_SVC, pc, cpsr
	mov	r0, sp				@ save current stack into r0
						@ (param register)
	.endm

	.macro	irq_save_user_regs
	sub	sp, sp, #S_FRAME_SIZE
	stmia	sp, {r0 - r12}			@ Calling r0-r12
	add	r8, sp, #S_PC			@ !! R8 NEEDS to be saved !!
						@ a reserved stack spot would
						@ be good.
	stmdb	r8, {sp, lr}^			@ Calling SP, LR
	str	lr, [r8, #0]			@ Save calling PC
	mrs	r6, spsr
	str	r6, [r8, #4]			@ Save CPSR
	str	r0, [r8, #8]			@ Save OLD_R0
	mov	r0, sp
	.endm

	.macro	irq_restore_user_regs
	ldmia	sp, {r0 - lr}^			@ Calling r0 - lr
	mov	r0, r0
	ldr	lr, [sp, #S_PC]			@ Get PC
	add	sp, sp, #S_FRAME_SIZE
	subs	pc, lr, #4			@ return & move spsr_svc into
						@ cpsr
	.endm

	.macro get_bad_stack
	ldr	r13, IRQ_STACK_START_IN		@ setup our mode stack (enter
						@ in banked mode)

	str	lr, [r13]			@ save caller lr in position 0
						@ of saved stack
	mrs	lr, spsr			@ get the spsr
	str	lr, [r13, #4]			@ save spsr in position 1 of
						@ saved stack

	mov	r13, #MODE_SVC			@ prepare SVC-Mode
	@ msr	spsr_c, r13
	msr	spsr, r13			@ switch modes, make sure
						@ moves will execute
	mov	lr, pc				@ capture return pc
	movs	pc, lr				@ jump to next instruction &
						@ switch modes.
	.endm

	.macro get_bad_stack_swi
	sub	r13, r13, #4			@ space on current stack for
						@ scratch reg.
	str	r0, [r13]			@ save R0's value.
	ldr	r0, IRQ_STACK_START_IN		@ get data regions start
						@ spots for abort stack
	str	lr, [r0]			@ save caller lr in position 0
						@ of saved stack
	mrs	r0, spsr			@ get the spsr
	str	lr, [r0, #4]			@ save spsr in position 1 of
						@ saved stack
	ldr	r0, [r13]			@ restore r0
	add	r13, r13, #4			@ pop stack entry
	.endm

	.macro get_irq_stack			@ setup IRQ stack
	ldr	sp, IRQ_STACK_START
	.endm

	.macro get_fiq_stack			@ setup FIQ stack
	ldr	sp, FIQ_STACK_START
	.endm

/*
 * exception handlers
 */
	.align	5
undefined_instruction:
	get_bad_stack
	bad_save_user_regs
	bl	do_undefined_instruction

	.align	5
software_interrupt:
	get_bad_stack_swi
	bad_save_user_regs
	bl	do_software_interrupt

	.align	5
prefetch_abort:
	get_bad_stack
	bad_save_user_regs
	bl	do_prefetch_abort

	.align	5
data_abort:
	get_bad_stack
	bad_save_user_regs
	bl	do_data_abort

	.align	5
not_used:
	get_bad_stack
	bad_save_user_regs
	bl	do_not_used

#ifdef CONFIG_USE_IRQ

	.align	5
irq:
	get_irq_stack
	irq_save_user_regs
	bl	do_irq
	irq_restore_user_regs

	.align	5
fiq:
	get_fiq_stack
	/* someone ought to write a more effective fiq_save_user_regs */
	irq_save_user_regs
	bl	do_fiq
	irq_restore_user_regs

#else

	.align	5
irq:
	get_bad_stack
	bad_save_user_regs
	bl	do_irq

	.align	5
fiq:
	get_bad_stack
	bad_save_user_regs
	bl	do_fiq

#endif /* CONFIG_USE_IRQ */
#endif /* CONFIG_SPL_BUILD */
